package Trans;

use strict;
use warnings;

=head1 trans.pm (i18n)

Povides enhanced translation routine.

=cut

use vars qw(%dictionary @EXPORT_OK);

@Trans::ISA = ('Exporter');

@EXPORT_OK = ('translate');

$Trans::VERSION = '0.1.0';
$Trans::NAME = 'Trans';

# contains dictionary copied from language module;
%dictionary=();

print "\ni18n/Trans.pm ($Trans::NAME $Trans::VERSION) loaded. Returning."
	if ($::DEBUG & 0x01);

=head1 FUNCTIONS

=cut

=head2 setup ($LANG)

Loads dictionary according to passed $LANG or $ENV{LANG}.
Dictionary is looked up in @INC as $LANG.pm.

Requires en.pm to be present - default-to-clean-english
does not work. English messages need to be 'translated',
because i moved all data labels into dictionary. And some
default dictionary should be loaded, if native will miss
some newly added messages or datalabels.

Entries from native dictionary replaces english.

=cut

sub setup
{
    my $lang = shift;
    $lang = $ENV{'LANG'} unless( $lang );

    $lang =~ s/_\w+//;
    
    # load default dictionary
    require "en.pm";
    %dictionary = %Language::dict;
    %Language::dict=();

    # replace entries from native
    if( grep { -f "$_/$lang.pm" } @INC ) {
	require "$lang.pm";
	print $/, 'i18n/', $Trans::NAME, ": loaded dictionary $lang.pm"
		if $::DEBUG;
	%dictionary = (%dictionary, %Language::dict);
	%Language::dict=();
	}
    else {
	print STDERR $/, 'i18n/', $Trans::NAME,
		": warning: dictionary for language '$lang' not found. ",
		"Falling back to 'POSIX'";
	}
}

=head2 translate ($FORMAT, @VALUES)

Uses FORMAT and VALUES like sprintf, but translates FORMAT and each of ARG
using current dictionary.

First, it looks up FORMAT in dictionary and uses it,
or falls back to originally passed;
Then handles expansions that modify message text;
Then handles expansions that modify argument values;
Then looks up each argument in dictionary and insert it in place of %s;

Modifying argument values:

 %[TEXT] replaces corresponding argument with value of TEXT
 %{EXPR} replaces with value of perl-evaluated EXPR

Idea of them are to chose proper form of word (argument value)
in depend on context.

Inside TEXT and EXPR occurences of '&','$N','$#' are replaced with value
of corresponding arg, Nth arg or number of arguments passed.
Escaping them with '\' works well.

Modifying message text:

 /TEXT/JOIN/	replaced with repeated TEXT, joined with JOIN,
                to make number of '%s' in message fit number of
		actual arguments; TEXT may contain %[] and %{};

Idea of /TEXT/JOIN/ is to handle messages with variable number of arguments.

 ${EXPR}	replaced with value of perl-evaluated expr,
		may contain all other expansions, except recursion.

Idea of ${EXPR} is to insert proper pronouns or whatever in depend on
argument values or number.

=cut

# FIXME: EXPR cannot contain '}' because of simple matching.
#        Looking for matching } should be possible with (?{EXPR}), but
#        it seems they have the same problems with closing brace.
#        and it will be abit slower.

sub translate
{
    my ($message,@args)=@_;
    my ($head,$expr);
    my ($rest,@sargs);

    return undef unless(defined $message);

    # preserve originally given format and args to fallback
    my ($fmessage,@fargs)=($message,@args);

    ### lookup translated formatstring
    $message = $dictionary{$message};
    $message = $fmessage unless ($message);

    print STDERR "\n$fmessage -> $message" if( $::DEBUG&128 );

    ### ${EXPR}
    $rest=$message; $message="";
    for(my $cnt=0;$rest=~s/^(.*?)\$\{(.*?)\}//s;$cnt++) {
	($head,$expr)=($1,$2);
	$expr=~s/(?<!\\)\$#/scalar(@args)/ge;
	$expr=~s/(?<!\\)\$(\d)/$args[$1-1]/ge;
	$expr=eval($expr);
	$message.="$head$expr";
	}
    $message.=$rest;

    print STDERR " -> $message" if( $::DEBUG&128 );

    ### /TEXT/JOIN/
    $rest=$message;
    if( $rest =~ s/^(.*?)\/(%([sd]|\[.*?\]|\{.*?\}).*?)\/(.*?)\///is ) {
	my ($head,$expr,$join) = ($1,$2,$4);
	my ($num,$cnt);
	$message = $head;
	# count number of %
	$num = ($head=~s/%/%/g);
	$num += ($rest=~s/%/%/g);
	$num = (scalar(@args)-$num);
	$cnt = ($expr=~s/%/%/g);
	$num /= $cnt;
	$message .= join($join, ($expr) x $num) if( $num > 0 );
	$message .= $rest;
	}

    print STDERR " -> $message (".join(',',@args).")" if( $::DEBUG&128 );

    ### %{EXPR} and %[TEXT]
    $rest=$message; $message="";
    @sargs=@args;
    for(my $cnt=0; $rest=~s/^(.*?)%([sd]|\[.*?\]|\{.*?\})//is; $cnt++) {
	($head,$expr)=($1,$2);
	if( $expr =~ /^[sd]$/ ) { $message.="$head%$expr"; next; }
	
	$expr=~s/(?<!\\)\$#/scalar(@args)/ge;
	$expr=~s/(?<!\\)\$(\d)/$sargs[$1-1]/ge;
	$expr=~s/(?<!\\)&/$sargs[$cnt]/ge;
	
	if( ($expr =~ s/^\{(.*)\}$/$1/s) ) {
	    $expr=eval($expr);
	    }
	elsif( $expr =~ s/^\[(.*)\]$/$1/s ) {
	    $expr=~s/\\&/&/g;
	    $expr=~s/\\\$/\$/g;
	    }

	$args[$cnt]=$expr;
	$message.="$head%s";
	}
    $message.=$rest;

    print STDERR " -> $message (".join(',',@args).")" if( $::DEBUG&128 );

    ### lookup args
    # take original unmodified, if failed
    foreach my $cnt (0..$#args) {
	my $arg = $dictionary{$args[$cnt]};
	$arg = $fargs[$cnt] unless ($arg);
	$args[$cnt]=$arg if($arg);
	}

    print STDERR " -> splintf($message,".join(',',@args).")" if( $::DEBUG&128 );

    $message =~ s/%S/%s/g;
    return sprintf($message,@args);
}

=head1 AUTHOR

QMax E<lt>qmax@mail.ruE<gt>

=cut

1;
